//
#include "stdafx.h"

//
#include "PIC16xCore.h"



// ****************************************************************************
//
//                 ** CPIC16xCore CLASS /Ҹ **
//
// ****************************************************************************


//
// CPIC16xCore 
//
//  - 2002.7.10, skr
//
CPIC16xCore::CPIC16xCore(u16 szMemProgram,	// Program Memory ũ
						 u16 szMemEeprom)	// EEPROM Memory ũ

	: m_MemProgramSize(szMemProgram), m_MemEepromSize(szMemEeprom)

{
#ifdef _VPM_CONSOLE_TEST
	printf("CPIC16xCore Created!! - Program Memory Size : %d.\n", m_MemProgramSize);
#endif

	//
	//  ޸ Ҵ
	//
	m_pMemProgram	= new u16[szMemProgram];
	m_pMemData		= new u8[NUMBER_OF_DATA_MEMORY_BANKS * DATA_MEMORY_BANK_SIZE];
	m_pMemEeprom	= new u8[szMemEeprom];

	//  ޸𸮰  Ҵ ƴ Ȯ
	if(!m_pMemProgram || !m_pMemEeprom || !m_pMemData)
	{
		m_bErrorOccurred = true;	// ȵƴٸ  üũ

#ifdef _VPM_CONSOLE_TEST
		printf("Core Memory Allocation FAILURE!!\n");
#endif
	}
	else
	{
		m_bErrorOccurred = false;

#ifdef _VPM_CONSOLE_TEST
		printf("Core Memory Allocation SUCCESS!!\n");
#endif
	}

	//
	// Ŭ ϴ   ʱȭ
	//
	m_ClkQ = Q1;
	m_pOpFunc = NULL;

	//
	m_pMemProgram[0] = 0x30ff; // 11 00xx 1111 1110 : MOVLW 0FEh
	m_pMemProgram[1] = 0x3f02; // 11 111x 0000 0010 : ADDLW 2
	m_pMemProgram[2] = 0x0114; // 00 0001 0001 0100 : CLRW
}

//
// CPIC16xCore Ҹ
//
//  - 2002.7.19, skr
//
CPIC16xCore::~CPIC16xCore(void)
{
#ifdef _VPM_CONSOLE_TEST
	printf("CPIC16xCore Destroyed!\n");
#endif

	//
	//  ޸ 
	//
	if(m_pMemProgram)	{ delete []m_pMemProgram;		m_pMemProgram = NULL;	}
	if(m_pMemData)		{ delete []m_pMemData;			m_pMemData = NULL;		}
	if(m_pMemEeprom)	{ delete []m_pMemEeprom;		m_pMemEeprom = NULL;	}
}



// ****************************************************************************
//
//                 ** DEVICE CLASS  ִ Լ **
//
// ****************************************************************************


//
// ھ  ִ ˷ִ Լ
//
// - return : m_bErrorOccurred;
//
//  - 2002.7.14, skr
//
bool
CPIC16xCore::core_is_error_occurred(void)
const
{
	return (bool)(m_bErrorOccurred);
}

//
// α׷ ޸ ͸  Լ
//
// - return : α׷ ޸  
//
//  - 2002.7.10, skr
//
u16*
CPIC16xCore::core_get_program_mem_ptr(void)
const
{
	return (u16*)(m_pMemProgram);
}

//
//  ޸ ͸  Լ
//
// - return :  ޸ 0  
//
//  - 2002.7.10, skr
//
u8*
CPIC16xCore::core_get_data_mem_ptr(void)
const
{
	return (u8*)(m_pMemData);
}

//
// EEPROM ޸ ͸  Լ
//
// - return : eeprom ޸  
//
//  - 2002.7.10, skr
//
u8*
CPIC16xCore::core_get_eeprom_mem_ptr(void)
const
{
	return (u8*)(m_pMemEeprom);
}

//
// PC(Program Counter)  ͸  Լ
//
// - return : PC  
//
//  - 2002.7.10, skr
//
u16
CPIC16xCore::core_get_pc_reg(void)
const
{
	return (u16)( ((u16)(m_pMemData[0x0a] << 8)) | ((u16)(m_pMemData[0x02])) );
}

//
// Stack  ͸  Լ
//
// - return : stack  0° (ֻ ) 
//
//  - 2002.7.10, skr
//
const u16*
CPIC16xCore::core_get_stack_reg_ptr(void)
const
{
	return (u16*)(&m_RegStack[0]);
}

//
// Instruction  ͸  Լ
//
// - return : instruction  
//
//  - 2002.7.10, skr
//
u16
CPIC16xCore::core_get_instruction_reg(void)
const
{
	return (u16)(m_RegInstruction);
}

//
// Working(W)  ͸  Լ
//
// return - w  
//
//  - 2002.7.10, skr
//
u8
CPIC16xCore::core_get_working_reg(void)
const
{
	return (u8)(m_RegWorking);
}



// ****************************************************************************
//
//                ** DEVICE CLASSκ  Է¹޴ Լ **
//
// ****************************************************************************


//
// ھ ̴   ʱȭ Ѵ.
//  Power-on Reset, /MCLR Reset Ǿ ִ.
//
// - return : m_bErrorOccurred  Ѵ.
//
//  - 2002.7.13, skr
//
bool
CPIC16xCore::core_init(INIT_CON InitCondition)
{
	//  ޸ Ҵ  ̷  ʾҴٸ
	//  Ѵ.
	if(m_bErrorOccurred) {
		return (bool)(false);
	}

	//
	switch(InitCondition)
	{

		// Power-on Reset
	case RST_POR_CON:
		{
			// register
			m_RegWorking		= 0;	// Init W Register
			m_RegStackPointer	= -1;	// Init Stack Count

			// Data memory reset
			memset(
				(void *)m_pMemData, 
				0x00, 
				((NUMBER_OF_DATA_MEMORY_BANKS * DATA_MEMORY_BANK_SIZE) * sizeof(u8))
			);

			// set data memory
			m_pMemData[0x03] = 0x18;	// STATUS (0001 1000)

		}
		break;

		// /MCLR Reset
	case RST_MCLR_CON:
		{
			// register
			m_RegStackPointer = -1;	// Init Stack Count

			// set data memory
			SetPC(0x0000);											// PC (0 0000 0000 0000)
			m_pMemData[0x03]  = ((m_pMemData[0x03] | 0x18) & 0x1f);	// STATUS (0001 1uuu)
			m_pMemData[0x0b] &= 0x01;								// INTCON
		}
		break;

	} // switch(InitCondition)

	return (bool)(true);
}


//
// ھ Ŭ ȣ Է ޴ Լ
//
// - clock : CLK_HIGH or CLK_LOW
// - return :  Q Clock ó Լ Ų 
//
//  - 2002.7.10, skr
//
bool
CPIC16xCore::core_clock_in(CLK_SIGNAL clock)
{
	// ھ  ޸ Ҵ翡 ߾ٸ  
	if(m_bErrorOccurred) {
		return (bool)(false);
	}

	// clock low ƹ ϵ  ʴ´.
	if(clock == CLK_LOW) {
		return (bool)(true);
	}

	//
	bool bResult = false;

	//  Q Clock ´ ۾ Լ ȣѴ.
	//  m_ClkQ   Q Clock ű.
	switch(m_ClkQ)
	{
	case Q1:	if((bResult = ExcuteQ1Cycle()))	m_ClkQ = Q2;	break;
	case Q2:	if((bResult = ExcuteQ2Cycle()))	m_ClkQ = Q3;	break;
	case Q3:	if((bResult = ExcuteQ3Cycle()))	m_ClkQ = Q4;	break;
	case Q4:	if((bResult = ExcuteQ4Cycle()))	m_ClkQ = Q1;	break;
	}

	return (bool)(bResult);
}



// ****************************************************************************
//
//             **  ھ ϴ   Լ **
//
// ****************************************************************************


//
// Q1 Clock ھ ϴ  Ѵ.
// ھ ̶ PC  1 Ų.
//
// - return : PC α׷ ޸   ʾҴٸ true .
//
//  - 2002.7.10, skr
//
bool
CPIC16xCore::ExcuteQ1Cycle(void)
{
	//  PC α׷ ޸  ߴٸ
	// ̻  ʴ´.
	if(core_get_pc_reg() == m_MemProgramSize) {
		return (bool)(false);
	}

	//
	u16 uResult;

	//  Fetch ɾ о´.
	uResult = Fetch(FETCH_GET);

	//  Fetch ɾ Branch(б) ɾ ƴ϶
	// PC Ų. 
	if(uResult != U16_MAX_VALUE) {
		IncPC();
	} 
	// б ̾ٸ ο PC  ɾ Fetchϱ 
	// NOPɾ ϰ 1 Cycle  Ѵ.
	else {
		uResult = 0x0000; // NOP Operation
	}

	// Instruction Register ɾ ԷѴ.
	m_RegInstruction = uResult;

	return (bool)(true);
}

//
// Q2 Clock ھ ϴ  Ѵ.
//
// - return : InstructionDecode  ° Ѵ.
//
//  - 2002.7.10, skr
//
bool
CPIC16xCore::ExcuteQ2Cycle(void)
{
	return (bool)(InstructionDecode());
}

//
// Q3 Clock ھ ϴ  Ѵ.
//
// - return : ɾ Ų  Ѵ.
//
//  - 2002.7.10, skr
//
bool
CPIC16xCore::ExcuteQ3Cycle(void)
{
	// Instruction ؼ Լ ã ߴٸ
	if(m_pOpFunc == NULL) {
		return (bool)(false);
	}

	// Instructoin ؼѴ.
	return (bool)((this->*m_pOpFunc)());
}

//
// Q4 Clock ھ ϴ  Ѵ.
//
// - return : ɾ Ų  Ѵ.
//
//  - 2002.7.10, skr
//
bool
CPIC16xCore::ExcuteQ4Cycle(void)
{
	if(m_pOpFunc == NULL) return false;

	return (bool)((this->*m_pOpFunc)());
}

//
// Stack  Pop Ѵ.
//
// - return :   ϰ ׻ true 
//
//  - 2002.7.11, skr
//
bool
CPIC16xCore::PopStack(void)
{
	// ÿ  PC   .
	if(m_RegStackPointer < 0) {
//		m_RegPC = 0;
		return (bool)(false);
	}

	// PC TOS(Top of the Stack)  Ѵ. 
	SetPC(m_RegStack[m_RegStackPointer]);

	//  Access Stack ġ Ѵ.
	m_RegStackPointer--;

	return (bool)(true);
}

//
// Stack  Push Ѵ.
//
// - return : Stack  á ϰ ׻ true 
//
//  - 2002.7.11, skr
//
bool
CPIC16xCore::PushStack(void)
{
#define STACK_SIZE		8
	//   áٸ  
	if(m_RegStackPointer >= (STACK_SIZE-1)) {
//		m_RegPC = 0;
		return (bool)(false);
	}
#undef STACK_SIZE

	// Access Stack ġ Ѵ.
	m_RegStackPointer++;

	//  TOS(Top of the Stack) PC Ѵ.
	m_RegStack[m_RegStackPointer] = core_get_pc_reg();
	
	return (bool)(true);
}

//
// PC(Program Counter)   
//
// - NewPC : ο PC 
// - reutrn : ׻ true
//
//  - 2002.7.13, skr
//
bool 
CPIC16xCore::SetPC(u16 NewPC)
{
	m_pMemData[0x02] = (u8)(NewPC);			// PCL
	m_pMemData[0x0a] = (u8)(NewPC >> 8);	// PCLATH

	return (bool)(true);
}

//
// PC(Program Counter)  1 Ų.
//
// - reutrn : ׻ true
//
//  - 2002.7.13, skr
//
bool
CPIC16xCore::IncPC(void)
{
	SetPC( (core_get_pc_reg() + 1) );

	return (bool)(true);
}

/*
----------------- FETCH Excution Ͽ, 7.11, skr ---------------

	- Clocking Scheme/Instruction Cycle -
1.  Q1 Clock  PC Ѵ.
2. Q4 Clock α׷ ޸𸮿 Instruction Fetchǰ 
  Instruction Register Latch ȴ.
3. Q1 ~ Q4 Clock  Instruction Decode ǰ Ǵ.

	- Instruction Flow/Pipelining -
1. Instruction Fetch  pipeline  Ǿ ִ.
2. Instruction Fetch 1 Instruction Cycle ʿ ϸ,
  Instruction Decode ϰ ϱ ؼ ٸ Instruction Cycle 䱸ȴ.
3. ׷ pipeline    1 Cycle ȿ Ǿ Ѵ.
4.  PC ȭŰ  Ϻϰ ϱ ؼ 2 Cycles ʿ Ѵ.
5. Fetch Cycle Q1 Clock PC  Ű Ѵ.
6. Excution Cycle  Fetch ɾ Q1 Clock Instruction Register Latchȴ.
   ɾ Q2, Q3, Q4 Clock Decodeǰ ȴ. Q2 Clock Data Memory
  о ̸, Q4 Clock  Ѵ.



 - Q1 Clock  -
1. FetchǾ ɾ "Instruction Register" Latch Ų.
2. PC  Ų. (PC = PC+1) - ʱ PC  0 ̴.

 - Q2 Clock  -
1. Excution Cycle ɾ ޸𸮸 ڵ Ѵ.

 - Q3 Clock  -
1. Excution Cycle LatchǾ Instruction Ѵ.

 - Q4 Clock  -
1.  ͳ ޸𸮿 Ѵ. (PC ̶ )
2. Fetch ¸ Ѵ.
3. Device  ֺġ  Ҽֵ  Լ ȣѴ.
*/

//
// Fetch   Լ
//  Լ Q1, Q4 Clock ѹ ȣǸ, Q1 Clock PC
// Ű  FETCH_GET Ķͷ, Q4 Clock  ޸ 
//  ģ б ̾ FETCH_BRANCH Ķͷ б ¸
// ϱ  Ѵ.
//
// - FetchCon : Լ ȣҶ Fetch Instruction   
// - return : Fetch ɾ ¸ Ѵ.
//
//  - 2002.7.15, skr
// 
u16
CPIC16xCore::Fetch(FETCH_CON FetchCon)
{
	static bool bBranchOccurred = false;

	if(FetchCon == FETCH_GET) {
		if(bBranchOccurred) {
			// Q1 Clock ó Լ Fetch FETCH_GET ĶͿ
			// Բ ȣ ϰ U16_MAX_VALUE  Branch
			// ־ٴ  ˰ PC  Ű  NOP 
			// Ų. 
			bBranchOccurred = false;

			return (u16)(U16_MAX_VALUE);
		}

		//  Fetch  Ѵ.
		return (u16)(m_pMemProgram[core_get_pc_reg()]);
	} else {
		// Branch Ű Լ ƴ ¸ Ѵ.
		bBranchOccurred = true;
	}

	// FetchCon FETCH_BRANCH ϶ FetchԼ  ƴٸ 0 ƴѰ
	// Ͽ   ˸.
	return (u16)(U16_MAX_VALUE);
}

/*
----------- PIC ɾ  , 7.11, skr ---------

xx xxxx xxxx xxxx
~~ ~~~~ ~~~~ ~~~~
3    2    1   0   - ° Ϻ


- 3° Ϻ 00϶ 2° Ϻ   ɾ  -
0:  6 - MOVWF, NOP, SLEEP, CLRWDT, RETURN, RETFIE
1:  2 - CLRW, CLRF
2:  1 - SUBWF
3:  1 - DECF
4:  1 - IORWF
5:  1 - ANDWF
6:  1 - XORWF
7:  1 - ADDWF
8:  1 - MOVF
9:  1 - COMF
10: 1 - INCF
11: 1 - DECFSZ
12: 1 - RRF
13: 1 - RLF
14: 1 - SWAPF
15: 1 - INCFSZ

- 3° Ϻ 00, 2° Ϻ 0000϶ ɾ и -
1. 1° Ϻ   Ʈ 1̸ MOVWF ɾ ̴.
2. 1° Ϻ   Ʈ 0̸ 5 ɾ 0° Ϻ  иؾ Ѵ.
0     : 1 - NOP
1~2   : 0
3     : 1 - SLEEP
4     : 1 - CLRWDT
5~7   : 0
8     : 1 - RETURN
9     : 1 - RETFIE
10~15 : 0

- 3° Ϻ 00, 2° Ϻ 0001϶ ɾ и -
- ̶ 1° Ϻ   Ʈ ɾ Ѵ. -
0 : 1 - CLRW
1 : 1 - CLRF

- 3° Ϻ 01϶ 2° Ϻ   ɾ  -
- ̶ 2° Ϻ(xxyy)  2Ʈ(yy)  ʴ´. -
-  2Ʈ(xx) Ͽ ɾ  -
0 : 1 - BCF
1 : 1 - BSF
2 : 1 - BTFSC
3 : 1 - BTFSS

- 3° Ϻ 10϶ 2° Ϻ   ɾ  -
- ̶ 2° Ϻ(xyyy)  3Ʈ(yyy)  ʴ´. -
-  1Ʈ(x) Ͽ ɾ  -
0 : 1 - CALL
1 : 1 - GOTO

- 3° Ϻ 11϶ 2° Ϻ   ɾ  -
0~3   : 1 - MOVLW
4~7   : 1 - RETLW
8     : 1 - IORLW
9     : 1 - ANDLW
10    : 1 - XORLW
11    : 0
12~13 : 1 - SUBLW
14~15 : 1 - ADDLW
*/

//
// Instruction Register ִ ɾ Ķ͸ ؼϴ Լ̴.
// Q2 Clock  ȴ. PIC16x ɾ   ڷḦ
//  .
//
// - return : bSuccess Ѵ.
//
//  - 2002.7.12, skr
//
bool
CPIC16xCore::InstructionDecode(void)
{
	bool bSuccess = true;
	m_pOpFunc = NULL;

	// Instruction Decoding...
	switch(((m_RegInstruction & 0x3fff) >> 12))
	{
		//
		// Ʈ   Ʈ ɾ ׷
		//
	case 0:
		{
			// ɾ ˸´ Լ  
			switch( ((m_RegInstruction & 0x0f00) >> 8) )
			{

			case 0:
				{
					if( (m_RegInstruction & 0x0080) ) {
						m_pOpFunc = Op_MOVWF;
					} else {
						switch( (m_RegInstruction & 0x000f) )
						{
						case 0:		m_pOpFunc = Op_NOP;		break;
						case 3:		m_pOpFunc = Op_SLEEP;	break;
						case 4:		m_pOpFunc = Op_CLRWDT;	break;
						case 8:		m_pOpFunc = Op_RETURN;	break;
						case 9:		m_pOpFunc = Op_RETFIE;	break;
						default:	bSuccess  = false;		break;
						}
					}
				}
				break;

			case 1:
				{
					if( (m_RegInstruction & 0x0080) ) {
						m_pOpFunc = Op_CLRF;
					} else {
						m_pOpFunc = Op_CLRW;
					}
				}
				break;

			case 2:		m_pOpFunc = Op_SUBWF;	break;
			case 3:		m_pOpFunc = Op_DECF;	break;
			case 4:		m_pOpFunc = Op_IORWF;	break;
			case 5:		m_pOpFunc = Op_ANDWF;	break;
			case 6:		m_pOpFunc = Op_XORWF;	break;
			case 7:		m_pOpFunc = Op_ADDWF;	break;
			case 8:		m_pOpFunc = Op_MOVF;	break;
			case 9:		m_pOpFunc = Op_COMF;	break;
			case 10:	m_pOpFunc = Op_INCF;	break;
			case 11:	m_pOpFunc = Op_DECFSZ;	break;
			case 12:	m_pOpFunc = Op_RRF;		break;
			case 13:	m_pOpFunc = Op_RLF;		break;
			case 14:	m_pOpFunc = Op_SWAPF;	break;
			case 15:	m_pOpFunc = Op_INCFSZ;	break;
			}

			// ޸ ڵ
			bSuccess = GetFRV(
				(u8)(m_RegInstruction & 0x007f), 
				&m_OpDecodeValue 
				);
		}
		break;

		//
		// Ʈ  ɾ ׷
		//
	case 1:
		{
			// ɾ ˸´ Լ  
			switch( ((m_RegInstruction & 0x0c00) >> 10) )
			{
			case 0:	m_pOpFunc = Op_BCF;		break;
			case 1: m_pOpFunc = Op_BSF;		break;
			case 2:	m_pOpFunc = Op_BTFSC;	break;
			case 3:	m_pOpFunc = Op_BTFSS;	break;
			}

			// ޸ ڵ
			bSuccess = GetFRV(
				(u8)(m_RegInstruction & 0x007f),
				&m_OpDecodeValue
				);
		}
		break;

		//
		// б ɾ ׷
		//
	case 2:
		{
			// ɾ ˸´ Լ  
			if( (m_RegInstruction & 0x0800) )	m_pOpFunc = Op_GOTO;
			else								m_pOpFunc = Op_CALL;
		}
		break;

		//
		// (Literal) ó ɾ ׷
		//
	case 3:
		{
			// ɾ ˸´ Լ  
			switch( ((m_RegInstruction & 0x0f00) >> 8) )
			{
			case 0:	case 1:	case 2:	case 3:		m_pOpFunc = Op_MOVLW;	break;
			case 4:	case 5:	case 6:	case 7:		m_pOpFunc = Op_RETLW;	break;
			case 8:								m_pOpFunc = Op_IORLW;	break;
			case 9:								m_pOpFunc = Op_ANDLW;	break;
			case 10:							m_pOpFunc = Op_XORLW;	break;
			case 12:	case 13:				m_pOpFunc = Op_SUBLW;	break;
			case 14:	case 15:				m_pOpFunc = Op_ADDLW;	break;
			default:							bSuccess = false;		break;
			}

			// (Literal) ڵ
			m_OpDecodeValue = (u8)(m_RegInstruction & 0x00ff);
		}
		break;

	} // switch(((m_RegInstruction & 0x3fff) >> 12))

	return (bool)(bSuccess);
}



// ****************************************************************************
//
//             **  ھο   ϴ Լ **
//
// ****************************************************************************


//
// ADDWF ɾ óϴ Լ
//
// - Op Code : 00 0111 dfff ffff
// - Cycles : 1
// - Status Affected : C, DC, Z
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.15, skr
//
bool
CPIC16xCore::Op_ADDWF(void)
{
	bool bSuccess = true;

	static u16 result;
	static u8  status;

	// Q3 Clock Instruction 
	if(m_ClkQ == Q3) {

		// status register ʱȭ
		status = 0;

		//  
		result = (u16)((u16)m_RegWorking + (u16)m_OpDecodeValue);

		// carry Ʈ 
		if( (result & 0x0100) ) {
			status |= 0x01; // 0000 000c
		}
		
		// digit carry Ʈ 
		if( ((m_RegWorking & 0x0f) + (m_OpDecodeValue & 0x0f)) > 0x0f ) {
			status |= 0x02; // 0000 00d0
		}

		// zero Ʈ 
		if( (result & 0x00ff) == 0 ) {
			status |= 0x04;	// 0000 0z00
		}

#ifdef _VPM_CONSOLE_TEST
	printf("ADDWF Instruction Excuted in Q3 Clock.\n");
#endif
	} 
	
	// Q4 Clock Instruction 
	else {

		// status 
		m_pMemData[0x03] &= 0xf8;	// 1111 1000
		m_pMemData[0x03] |= status;	// xxxx xZDC

		//  
		bSuccess = WriteOpResultToDest( (u8)(result & 0x00ff) );

#ifdef _VPM_CONSOLE_TEST
	printf("ADDWF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// ANDWF ɾ óϴ Լ
//
// - Op Code : 00 0101 dfff ffff
// - Cycles : 1
// - Status Affected : Z
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.15, skr
//
bool
CPIC16xCore::Op_ANDWF(void)
{
	bool bSuccess = true;

	static u8 result;
	static u8 status;

	if(m_ClkQ == Q3) {

		// status register ʱȭ
		status = 0;

		//  
		result = (m_RegWorking & m_OpDecodeValue);

		// zero flag  
		if(result == 0) {
			status = 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("ANDWF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// status 
		m_pMemData[0x03] &= 0xfb;	// 1111 1011
		m_pMemData[0x03] |= status;	// xxxx xZxx

		//  
		bSuccess = WriteOpResultToDest(result);

#ifdef _VPM_CONSOLE_TEST
	printf("ANDWF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// CLRF ɾ óϴ Լ
//
// - Op Code : 00 0001 1fff ffff
// - Cycles : 1
// - Status Affected : Z
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.15, skr
//
bool
CPIC16xCore::Op_CLRF(void)
{
	bool bSuccess = true;

	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("CLRF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		// setting zero flag
		m_pMemData[0x03] |= 0x04;

		//  
		bSuccess = WriteOpResultToDest( (u8)(0x00) );

#ifdef _VPM_CONSOLE_TEST
	printf("CLRF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// CLRW ɾ óϴ Լ
//
// - Op Code : 00 0001 0xxx xxxx
// - Cycles : 1
// - Status Affected : Z
// - return : ׻ true
//
//  - 2002.7.15, skr
//
bool
CPIC16xCore::Op_CLRW(void)
{
	bool bSuccess = true;

	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("CLRW Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		// setting zero flag
		m_pMemData[0x03] |= 0x04;

		//  
		bSuccess = WriteOpResultToDest( (u8)(0x00) );

#ifdef _VPM_CONSOLE_TEST
	printf("CLRW Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}

//
// COMF ɾ óϴ Լ
//
// - Op Code : 00 1001 dfff ffff
// - Cycles : 1
// - Status Affected : Z
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.15, skr
//
bool
CPIC16xCore::Op_COMF(void)
{
	bool bSuccess = true;

	static u8 result;
	static u8 status;

	if(m_ClkQ == Q3) {

		// status register ʱȭ
		status = 0;

		//  
		result = (u8)(~m_OpDecodeValue);

		// zero flag  
		if(result == 0) {
			status = 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("COMF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		// zero flag  
		m_pMemData[0x03] &= 0xfb;	// 1111 1011
		m_pMemData[0x03] |= status;	// xxxx xZxx

		//  
		bSuccess = WriteOpResultToDest(result);

#ifdef _VPM_CONSOLE_TEST
	printf("COMF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// DECF ɾ óϴ Լ
//
// - Op Code : 00 0011 dfff ffff
// - Cycles : 1
// - Status Affected : Z
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.15, skr
// 
bool
CPIC16xCore::Op_DECF(void)
{
	bool bSuccess = true;

	static u8 result;
	static u8 status;

	if(m_ClkQ == Q3) {

		// status register ʱȭ
		status = 0;

		//  
		result = (u8)(m_OpDecodeValue - 1);

		// zero flag  
		if(result == 0) {
			status = 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("DECF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		// status register 
		m_pMemData[0x03] &= 0xfb;	// 1111 1011
		m_pMemData[0x03] |= status;

		//  
		bSuccess = WriteOpResultToDest(result);

#ifdef _VPM_CONSOLE_TEST
	printf("DECF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// DECFSZ ɾ óϴ Լ
//
// - Op Code : 00 1011 dfff ffff
// - Cycles : 1(2)
// - Status Affected : none
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.15, skr
//
bool
CPIC16xCore::Op_DECFSZ(void)
{
	bool bSuccess = true;

	static u8 result;

	if(m_ClkQ == Q3) {

		//  
		result = (u8)(m_OpDecodeValue - 1);

#ifdef _VPM_CONSOLE_TEST
	printf("DECFSZ Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		//  
		bSuccess = WriteOpResultToDest(result);

		//  0̸ PC+1 бѴ.
		if( result == 0 ) {
			IncPC();				// PC 
			Fetch(FETCH_BRANCH);	// Branch  Ǿٴ° ˸.
		}

#ifdef _VPM_CONSOLE_TEST
	printf("DECFSZ Instruction Excuted in Q4 Clock. Result : %d\n", result);
#endif
	}

	return (bool)(bSuccess);
}

//
// INCF ɾ óϴ Լ
//
// - Op Code : 00 1010 dfff ffff
// - Cycles : 1
// - Status Affected : Z
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_INCF(void)
{
	bool bSuccess = true;

	static u8 result;
	static u8 status;

	if(m_ClkQ == Q3) {

		// status register ʱȭ
		status = 0;

		//  
		result = (u8)(m_OpDecodeValue + 1);

		// zero flag  
		if(result == 0) {
			status = 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("INCF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		// status register 
		m_pMemData[0x03] &= 0xfb;	// 1111 1011
		m_pMemData[0x03] |= status;

		//  
		bSuccess = WriteOpResultToDest(result);

#ifdef _VPM_CONSOLE_TEST
	printf("INCF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// INCFSZ ɾ óϴ Լ
//
// - Op Code : 00 1111 dfff ffff
// - Cycles : 1(2)
// - Status Affected : none
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_INCFSZ(void)
{
	bool bSuccess = true;

	static u8 result;

	if(m_ClkQ == Q3) {

		//  
		result = (u8)(m_OpDecodeValue + 1);

#ifdef _VPM_CONSOLE_TEST
	printf("INCFSZ Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		//  
		bSuccess = WriteOpResultToDest(result);

		//  0̸ PC+1 бѴ.
		if( result == 0 ) {
			IncPC();				// PC 
			Fetch(FETCH_BRANCH);	// Branch  Ǿٴ° ˸.
		}

#ifdef _VPM_CONSOLE_TEST
	printf("INCFSZ Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// IORWF ɾ óϴ Լ
//
// - Op Code : 00 0100 dfff ffff
// - Cycles : 1
// - Status Affected : Z
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_IORWF(void)
{
	bool bSuccess;

	static u8 result;
	static u8 status;

	if(m_ClkQ == Q3) {

		// status register ʱȭ
		status = 0;

		//  
		result = (m_RegWorking | m_OpDecodeValue);

		// zero flag  
		if(result == 0) {
			status = 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("IORWF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		// zero flag  
		m_pMemData[0x03] &= 0xfb;	// 1111 1011
		m_pMemData[0x03] |= status;

		//  
		bSuccess = WriteOpResultToDest(result);

#ifdef _VPM_CONSOLE_TEST
	printf("IORWF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// MOVF ɾ óϴ Լ
//
// - Op Code : 00 1000 dfff ffff
// - Cycles : 1
// - Status Affected : Z
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_MOVF(void)
{
	bool bSuccess = true;

	static u8 status;

	if(m_ClkQ == Q3) {

		// status register ʱȭ
		status = 0;

		// zero flag  
		if( m_OpDecodeValue == 0 ) {
			status = 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("MOVF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// zero flag  
		m_pMemData[0x03] &= 0xfb;	// 1111 1011
		m_pMemData[0x03] |= status;

		//  
		bSuccess = WriteOpResultToDest(m_OpDecodeValue);

#ifdef _VPM_CONSOLE_TEST
	printf("MOVF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// MOVWF ɾ óϴ Լ
//
// - Op Code : 00 0100 1fff ffff
// - Cycles : 1
// - Status Affected : none
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_MOVWF(void)
{
	bool bSuccess = true;

	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("MOVWF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		//  
		bSuccess = WriteOpResultToDest(m_RegWorking);

#ifdef _VPM_CONSOLE_TEST
	printf("MOVWF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// NOP ɾ óϴ Լ
//
// - Op Code : 00 0000 0xx0 0000
// - Cycles : 1
// - Status Affected : none
// - return : ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_NOP(void)
{
	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("NOP Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("NOP Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}

//
// RLF ɾ óϴ Լ
//
// - Op Code : 00 1101 dfff ffff
// - Cycles : 1
// - Status Affected : C (xxxx xxxC)
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_RLF(void)
{
	bool bSuccess = true;

	static u8 result;
	static u8 status;

	if(m_ClkQ == Q3) {

		// status register ʱȭ
		status = 0;
		
		// carry flag  
		if( (m_OpDecodeValue & 0x80) ) {
			status = 0x01;
		}

		// Left Rotate
		result  = (m_OpDecodeValue << 1);
		result |= (m_pMemData[0x03] & 0x01);	// 0000 0001

#ifdef _VPM_CONSOLE_TEST
	printf("RLF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		// carry flag  
		m_pMemData[0x03] &= 0xfe;	// 1111 1110
		m_pMemData[0x03] |= status;

		//  
		bSuccess = WriteOpResultToDest(result);

#ifdef _VPM_CONSOLE_TEST
	printf("RLF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// RRF ɾ óϴ Լ
//
// - Op Code : 00 1100 dfff ffff
// - Cycles : 1
// - Status Affected : C
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_RRF(void)
{
	bool bSuccess = true;

	static u8 status;
	static u8 result;

	if(m_ClkQ == Q3) {

		// status reigster ʱȭ
		status = 0;

		// carry flag  
		if( (m_OpDecodeValue & 0x01) ) {
			status = 0x01;
		}

		// Right Rotate
		result  = (m_OpDecodeValue >> 1);
		result |= ((m_pMemData[0x03] & 0x01) << 7);

#ifdef _VPM_CONSOLE_TEST
	printf("RRF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// carry flag  
		m_pMemData[0x03] &= 0xfe;	// 1111 1110
		m_pMemData[0x03] |= status;

		//  
		bSuccess = WriteOpResultToDest(result);

#ifdef _VPM_CONSOLE_TEST
	printf("RRF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// SUBWF ɾ óϴ Լ
//
// - Op Code : 00 0010 dfff ffff
// - Cycles : 1
// - Status Affected : C, DC, Z
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_SUBWF(void)
{
	bool bSuccess = true;

	static u8 status;
	static u8 result;

	if(m_ClkQ == Q3) {
		
		// status register ʱȭ
		status = 0;

		//   (SUBWF : f-w)
		if( m_OpDecodeValue >= m_RegWorking) {			// no borrow
			result = (m_OpDecodeValue - m_RegWorking);

			status |= 0x01; // borrow  carry Ʈ set ȴ.
		}
		else {											// borrow
			result = (m_RegWorking - m_OpDecodeValue);

//			status &= 0xfe; // borrow  carry Ʈ clear ȴ.
		}

		// Digit Carry Ʈ  
		if( (m_OpDecodeValue & 0x0f) >= (m_RegWorking & 0x0f) ) {
			status |= 0x02;	// digit borrow  digit carry Ʈ set ȴ.
		}

		// Zero Ʈ  
		if( result == 0 ) {
			status |= 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("SUBWF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// status register  
		m_pMemData[0x03] &= 0xf8;	// 1111 1000
		m_pMemData[0x03] |= status;

		//  
		bSuccess = WriteOpResultToDest(result);

#ifdef _VPM_CONSOLE_TEST
	printf("SUBWF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// SWAPF ɾ óϴ Լ
//
// - Op Code : 00 1110 dfff ffff
// - Cycles : 1
// - Status Affected : none
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_SWAPF(void)
{
	bool bSuccess = true;

	static u8 result;

	if(m_ClkQ == Q3) {
		
		//  
		result = ( ((m_OpDecodeValue & 0xf0) >> 4) | ((m_OpDecodeValue & 0x0f) << 4) );

#ifdef _VPM_CONSOLE_TEST
	printf("SWAPF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		//  
		bSuccess = WriteOpResultToDest(result);

#ifdef _VPM_CONSOLE_TEST
	printf("SWAPF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// XORWF ɾ óϴ Լ
//
// - Op Code : 00 0110 dfff ffff
// - Cycles : 1
// - Status Affected : Z
// - return : WriteOpResultToDest Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_XORWF(void)
{
	bool bSuccess = true;

	static u8 status;
	static u8 result;

	if(m_ClkQ == Q3) {
		
		// status register ʱȭ
		status = 0;

		//  
		result = (u8)(m_RegWorking ^ m_OpDecodeValue);

		// zero flag  
		if( result == 0 ) {
			status = 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("XORWF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// status register 
		m_pMemData[0x03] &= 0xfb;	// 1111 1011
		m_pMemData[0x03] |= status;

		//  
		bSuccess = WriteOpResultToDest(result);

#ifdef _VPM_CONSOLE_TEST
	printf("XORWF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// CLRWDT ɾ óϴ Լ
// - Op Code : 00 0000 0110 0100
// - Cycles : 1
// - Status Affected : /TO, /PD
//
//  - 2002.7.11, skr
//
bool
CPIC16xCore::Op_CLRWDT(void)
{
	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("CLRWDT Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("CLRWDT Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return true;
}

//
// RETFIE ɾ óϴ Լ
//
// - Op Code : 00 0000 0000 1001
// - Cycles : 2
// - Status Affected : none
// - return : PopStack Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_RETFIE(void)
{
	bool bSuccess = true;

	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("RETFIE Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// GIE (Global Interrupt Enable) Ʈ set
		m_pMemData[0x0b] |= 0x80;

		// TOS -> PC
		bSuccess = PopStack();

		// Fetch б  Ͼٴ° ˸.
		Fetch(FETCH_BRANCH);

#ifdef _VPM_CONSOLE_TEST
	printf("RETFIE Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// RETURN ɾ óϴ Լ
//
// - Op Code : 00 0000 0000 1000
// - Cycles : 2
// - Status Affected : none
// - return : PopStack Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_RETURN(void)
{
	bool bSuccess = true;

	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("RETURN Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		// TOS -> PC
		bSuccess = PopStack();

		// Fetch б  Ͼٴ° ˸.
		Fetch(FETCH_BRANCH);

#ifdef _VPM_CONSOLE_TEST
	printf("RETURN Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// SLEEP ɾ óϴ Լ
// - Op Code : 00 0000 0110 0011
// - Cycles : 1
// - Status Affected : /TO, /PD
//
//  - 2002.7.11, skr
//
bool
CPIC16xCore::Op_SLEEP(void)
{
	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("SLEEP Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("SLEEP Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return true;
}

//
// BCF ɾ óϴ Լ
//
// - Op Code : 01 00bb bfff ffff
// - Cycles : 1
// - Status Affected : none
// - return : SetFRV Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_BCF(void)
{
	bool bSuccess = true;

	static u8 result;

	if(m_ClkQ == Q3) {

		//
		u8 mask = 0x01;
		u8 bitnumber;
		
		//  
		bitnumber = (u8)((m_RegInstruction & 0x0380) >> 7);
		mask = (u8)(~(mask << bitnumber));
		result = (m_OpDecodeValue & mask);

#ifdef _VPM_CONSOLE_TEST
	printf("BCF Instruction Excuted in Q3 Clock."
		   "bit : %d, mask : %02x, result : %02x\n", bitnumber, mask, result);
#endif
	} else {
		
		//  
		bSuccess = SetFRV(
			(u8)(m_RegInstruction & 0x007f),	// address
			result								// value
			);

#ifdef _VPM_CONSOLE_TEST
	printf("BCF Instruction Excuted in Q4 Clock."
		   "result : %02x\n", result);
#endif
	}

	return (bool)(bSuccess);
}

//
// BSF ɾ óϴ Լ
//
// - Op Code : 01 01bb bfff ffff
// - Cycles : 1
// - Status Affected : none
// - return : SetFRV Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_BSF(void)
{
	bool bSuccess = true;

	static u8 result;

	if(m_ClkQ == Q3) {

		//
		u8 mask = 0x01;
		u8 bitnumber;
		
		//  
		bitnumber = (u8)((m_RegInstruction & 0x0380) >> 7);
		mask <<= bitnumber;
		result = (m_OpDecodeValue | mask);

#ifdef _VPM_CONSOLE_TEST
	printf("BSF Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		//  
		bSuccess = SetFRV(
			(u8)(m_RegInstruction & 0x007f),	// address
			result								// value
			);

#ifdef _VPM_CONSOLE_TEST
	printf("BSF Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// BTFSC ɾ óϴ Լ
//
// - Op Code : 01 10bb bfff ffff
// - Cycles : 1(2)
// - Status Affected : none
// - return : ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_BTFSC(void)
{
	static bool isSkip;

	if(m_ClkQ == Q3) {

		//
		u8 mask = 0x01;
		u8 bitnumber;

		// ° ʱȭ
		isSkip = false;

		//  
		bitnumber = (u8)((m_RegInstruction & 0x0380) >> 7);
		mask <<= bitnumber;

		if( (m_OpDecodeValue & mask) == 0 ) {
			isSkip = true;
		}
		
#ifdef _VPM_CONSOLE_TEST
	printf("BTFSC Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// °  б θ 
		if( isSkip == true ) {
			// PC 
			IncPC();
			// б ΰ ˸
			Fetch(FETCH_BRANCH);
		}

#ifdef _VPM_CONSOLE_TEST
	printf("BTFSC Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}

//
// BTFSS ɾ óϴ Լ
//
// - Op Code : 01 00bb bfff ffff
// - Cycles : 1(2)
// - Status Affected : none
// - return : ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_BTFSS(void)
{
	static bool isSkip;

	if(m_ClkQ == Q3) {

		//
		u8 mask = 0x01;
		u8 bitnumber;

		// ° ʱȭ
		isSkip = false;

		//  
		bitnumber = (u8)((m_RegInstruction & 0x0380) >> 7);
		mask <<= bitnumber;

		if( (m_OpDecodeValue & mask) ) {
			isSkip = true;
		}
		
#ifdef _VPM_CONSOLE_TEST
	printf("BTFSS Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// °  б θ 
		if( isSkip == true ) {
			// PC 
			IncPC();
			// б ΰ ˸
			Fetch(FETCH_BRANCH);
		}

#ifdef _VPM_CONSOLE_TEST
	printf("BTFSS Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}

//
// ADDLW ɾ óϴ Լ
//
// - Op Code : 11 111x kkkk kkkk
// - Cycles : 1
// - Status Affected : C, DC, Z
//
//  - 2002.7.11, skr
//
bool
CPIC16xCore::Op_ADDLW(void)
{
	static u16 result;
	static u8  status;

	if(m_ClkQ == Q3) {

		// status register ʱȭ
		status = 0;

		//  
		result = (u16)((u16)m_RegWorking + (u16)m_OpDecodeValue);

		// carry Ʈ 
		if( (result & 0x0100) ) {
			status |= 0x01; // 0000 000c
		}
		
		// digit carry Ʈ 
		if( ((m_RegWorking & 0x0f) + (m_OpDecodeValue & 0x0f)) > 0x0f ) {
			status |= 0x02; // 0000 00d0
		}

		// zero Ʈ 
		if( (result & 0x00ff) == 0 ) {
			status |= 0x04;	// 0000 0z00
		}

#ifdef _VPM_CONSOLE_TEST
	printf("ADDLW Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		// status 
		m_pMemData[0x03] &= 0xf8;	// 1111 1000
		m_pMemData[0x03] |= status;	// xxxx xZDC

		//  
		m_RegWorking = (u8)(result & 0x00ff);

#ifdef _VPM_CONSOLE_TEST
	printf("ADDLW Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}

//
// ANDLW ɾ óϴ Լ
//
// - Op Code : 11 1001 kkkk kkkk
// - Cycles : 1
// - Status Affected : Z
// - return : ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_ANDLW(void)
{
	static u8 status;
	static u8 result;

	if(m_ClkQ == Q3) {
		
		// status register ʱȭ
		status = 0;

		//  
		result = (m_RegWorking & m_OpDecodeValue);

		// zero flag  
		if(result == 0) {
			status = 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("ANDLW Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		// zero flag  
		m_pMemData[0x03] &= 0xfb;	// 1111 1011
		m_pMemData[0x03] |= status;

		//  
		m_RegWorking = result;

#ifdef _VPM_CONSOLE_TEST
	printf("ANDLW Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}

//
// IORLW ɾ óϴ Լ
//
// - Op Code : 11 1000 kkkk kkkk
// - Cycles : 1
// - Status Affected : Z
// - return : ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_IORLW(void)
{
	static u8 status;
	static u8 result;

	if(m_ClkQ == Q3) {
		
		// status register ʱȭ
		status = 0;

		//  
		result = (m_RegWorking | m_OpDecodeValue);

		// zero flag  
		if( result == 0 ) {
			status = 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("IORLW Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// zero flag  
		m_pMemData[0x03] &= 0xfb;	// 1111 1011
		m_pMemData[0x03] |= status;

		//  
		m_RegWorking = result;

#ifdef _VPM_CONSOLE_TEST
	printf("IORLW Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}

//
// MOVLW ɾ óϴ Լ
//
// - Op Code : 11 00xx kkkk kkkk
// - Cycles : 1
// - Status Affected : none
// - return : ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_MOVLW(void)
{
	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("MOVLW Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		m_RegWorking = m_OpDecodeValue;

#ifdef _VPM_CONSOLE_TEST
	printf("MOVLW Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}

//
// RETLW ɾ óϴ Լ
//
// - Op Code : 11 01xx kkkk kkkk
// - Cycles : 2
// - Status Affected : none
// - return : PopStack .
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_RETLW(void)
{
	bool bSuccess = true;

	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("RETLW Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// W Register  
		m_RegWorking = m_OpDecodeValue;

		// TOS -> PC
		bSuccess = PopStack();

		// б   ˸.
		Fetch(FETCH_BRANCH);

#ifdef _VPM_CONSOLE_TEST
	printf("RETLW Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// SUBLW ɾ óϴ Լ
//
// - Op Code : 11 110x kkkk kkkk
// - Cycles : 1
// - Status Affected : C, DC, Z
// - return : ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_SUBLW(void)
{
	static u8 status;
	static u8 result;

	if(m_ClkQ == Q3) {
		
		// status register ʱȭ
		status = 0;

		//   (k-w)
		if(m_OpDecodeValue >= m_RegWorking) {
			result = (m_OpDecodeValue - m_RegWorking);

			status |= 0x01;	// carry bit set - no borrow
		} else {
			result = (m_RegWorking - m_OpDecodeValue);

//			status &= 0xfe;	// carry bit clear - borrow
		}

		// digit carry  
		if( (m_OpDecodeValue & 0x0f) >= (m_RegWorking & 0x0f) ) {
			status |= 0x02;	// digit carry bit set - no borrow
		}

		// zero flag bit  
		if( result == 0 ) {
			status |= 0x04; // zero flag bit set
		}

#ifdef _VPM_CONSOLE_TEST
	printf("SUBLW Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// status register 
		m_pMemData[0x03] &= 0xf8;	// 1111 1000
		m_pMemData[0x03] |= status;

		//  
		m_RegWorking = result;

#ifdef _VPM_CONSOLE_TEST
	printf("SUBLW Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}

//
// XORLW ɾ óϴ Լ
//
// - Op Code : 11 1010 kkkk kkkk
// - Cycles : 1
// - Status Affected : Z
// - return : ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_XORLW(void)
{
	static u8 status;
	static u8 result;

	if(m_ClkQ == Q3) {
		
		// status register ʱȭ
		status = 0;

		//  
		result = (m_RegWorking ^ m_OpDecodeValue);

		// zero flag  
		if( result == 0 ) {
			status = 0x04;
		}

#ifdef _VPM_CONSOLE_TEST
	printf("XORLW Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// zero flag  
		m_pMemData[0x03] &= 0xfb;	// 1111 1011
		m_pMemData[0x03] |= status;

		//  
		m_RegWorking = result;

#ifdef _VPM_CONSOLE_TEST
	printf("XORLW Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}

//
// CALL ɾ óϴ Լ
//
// - Op Code : 10 0kkk kkkk kkkk
// - Cycles : 2
// - Status Affected : none
// - return : PushStack Լ ȣ 
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_CALL(void)
{
	bool bSuccess = true;

	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("CALL Instruction Excuted in Q3 Clock.\n");
#endif
	} else {

		//   PC STACK 
		bSuccess = PushStack();
		
		// ο PC 
		SetPC( (u16)(m_RegInstruction & 0x03ff) );

		// б  Ͼٴ° 
		Fetch(FETCH_BRANCH);

#ifdef _VPM_CONSOLE_TEST
	printf("CALL Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(bSuccess);
}

//
// GOTO ɾ óϴ Լ
//
// - Op Code : 10 1kkk kkkk kkkk
// - Cycles : 2
// - Status Affected : none
// - return : ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::Op_GOTO(void)
{
	if(m_ClkQ == Q3) {
		;

#ifdef _VPM_CONSOLE_TEST
	printf("GOTO Instruction Excuted in Q3 Clock.\n");
#endif
	} else {
		
		// ο PC 
		SetPC( (u16)(m_RegInstruction & 0x03ff) );

		// б  Ͼٴ° 
		Fetch(FETCH_BRANCH);

#ifdef _VPM_CONSOLE_TEST
	printf("GOTO Instruction Excuted in Q4 Clock.\n");
#endif
	}

	return (bool)(true);
}



// ****************************************************************************
//
//             ** CPIC16xCore Class  ϴ Լ **
//
// ****************************************************************************


//
// Data Memory  д´.
//
// - RegAddr :  7Ʈ ּҰ
// - pValue :   
// - return : ޸ Ҵ翡 Ѱ͸   true Ѵ.
//            ׿ device class óؾ ҰͿ ؼ ˼.
//
//  - 2002.7.14, skr
//
bool
CPIC16xCore::GetFRV(u8 RegAddr, 
					u8 *pValue)
{
	// ޸ Ҵ ߴٸ  
	if(m_bErrorOccurred == true) {
		return (bool)(false);
	}

	u8 bank;
	bool IsINDF = false;

	// INDF   ȮѴ.
	if(RegAddr == 0x00) {
		bank = ( 
			((m_pMemData[0x03] & 0x80) >> 6)	// STATUS  IRP 
			| ((m_pMemData[0x04] >> 7) & 0x01)	// FSR  MSB 
			);

		RegAddr = (m_pMemData[0x04] & 0x7f);	// ο ּҸ οѴ.

		IsINDF = true;	// INDF ٴ° ˷ش.
	} 
	else {
		bank = (
			((m_pMemData[0x03] & 0x60) >> 5)	// bank 
			);
	}


	// GetFRV Լ Core   ͸ ϰ 
	// Device Class ӵ core_get_frv_ext Լ óѴ.
	//   ġ ϴ ޸  ٸ ̴.
	switch(RegAddr)
	{
	case 0x02: // PCL
	case 0x03: // STATUS
	case 0x04: // FSR
	case 0x0b: // INTCON
		{
			*pValue = m_pMemData[RegAddr];
		}
		break;

	case 0x0a: // PCLATH
		{
			if(IsINDF == true) {				// PCLATH INDF ϶ бⰡ ϴ.
				*pValue = m_pMemData[RegAddr];
			}
		}
		break;

	default: //  core_get_frv_ext Լ ó
		return (bool)(core_get_frv_ext(bank, RegAddr, pValue));	// device class ǵǾ ִ.
	}

	return (bool)(true);
}

//
// Data Memory  Ѵ.
//
// - RegAddr :  7Ʈ ּҰ
// - Value :  
// - return : ޸ Ҵ翡 Ѱ͸   true Ѵ.
//            ׿ device class óؾ ҰͿ ؼ ˼.
//
//  - 2002.7.14, skr
//
bool
CPIC16xCore::SetFRV(u8 RegAddr, u8 Value)
{
	// ޸ Ҵ ߴٸ  
	if(m_bErrorOccurred == true) {
		return (bool)(false);
	}

	u8 bank;
	bool IsINDF = false;

	// INDF   ȮѴ.
	if(RegAddr == 0x00) {
		bank = ( 
			((m_pMemData[0x03] & 0x80) >> 6)	// STATUS  IRP 
			| ((m_pMemData[0x04] >> 7) & 0x01)	// FSR  MSB 
			);

		RegAddr = (m_pMemData[0x04] & 0x7f);	// ο ּҸ οѴ.

		IsINDF = true;	// INDF ٴ° ˷ش.
	} 
	else {
		bank = (
			((m_pMemData[0x03] & 0x60) >> 5)	// bank 
			);
	}


	// SetFRV Լ Core   ͸ ϰ 
	// Device Class ӵ core_set_frv_ext Լ óѴ.
	//   ġ ϴ ޸  ٸ ̴.
	switch(RegAddr)
	{
	case 0x02: // PCL
	case 0x03: // STATUS
	case 0x04: // FSR
	case 0x0b: // INTCON
		{
			m_pMemData[RegAddr] = Value;
		}
		break;

	case 0x0a: // PCLATH
		{
			if(IsINDF == true) {				// PCLATH INDF ϶ Ⱑ ϴ.
				m_pMemData[RegAddr] = Value;
			}
		}
		break;

	default: //  core_set_frv_ext Լ ó
		return (bool)(core_set_frv_ext(bank, RegAddr, Value));	// device class ǵǾ ִ.
	}

	return (bool)(true);
}

//
// Byte-Oriented ɿ destination f, w  
// ̸  óϴ Լ̴.
//
// - result :  
// - return : SetFRV Լ   ϰ ׻ true
//
//  - 2002.7.16, skr
//
bool
CPIC16xCore::WriteOpResultToDest(u8 result)
{
	bool bSuccess = true;

		// file register  
		if( (m_RegInstruction & 0x0080) ) {
			bSuccess = SetFRV( 
				(u8)(m_RegInstruction & 0x007f),	// address
				result								// value
				);
		}

		// working register  
		else {
			m_RegWorking = result;
		}

	return (bool)(bSuccess);
}