//
#include "stdafx.h"

#include "VpmDefinition.h"

//
#ifndef _PIC16X_CORE
#define _PIC16X_CORE

//
#define NUMBER_OF_DATA_MEMORY_BANKS		4
#define DATA_MEMORY_BANK_SIZE			128

//
enum INIT_CON { RST_POR_CON,
				RST_MCLR_CON,	/* = WDT Reset */ 
				RST_WDT_CON,
				WKU_WDT_CON,	/* wake-up via WDT or Interrupt */
				WKU_INT_CON
};

//
static enum Q_CLOCKS  { Q1, Q2, Q3, Q4 };
static enum FETCH_CON { FETCH_GET, FETCH_BRANCH };

//
class CPIC16xCore
{
protected:
	CPIC16xCore(u16, u16);	// 
	~CPIC16xCore(void);		// Ҹ

	//
	// Device Class  ִ Լ
	//
protected:
	// ھ  
	bool core_is_error_occurred			(void) const;	// ھ  ִ 

	// ޸  
	u16* core_get_program_mem_ptr		(void) const;	// α׷ ޸ 
	u8*  core_get_data_mem_ptr			(void) const;	//  ޸ 
	u8*  core_get_eeprom_mem_ptr		(void) const;	// EEPROM ޸ 

	//    - pointer
	const u16* core_get_stack_reg_ptr	(void) const;	//   

	//    - value
	u16 core_get_pc_reg					(void) const;	// PC  
	u16 core_get_instruction_reg		(void) const;	// νƮ  
	u8  core_get_working_reg			(void) const;	// W  

	//
	// Device Class  Է¹޴ Լ
	//
protected:
	// core class ʱȭ ϴ Լ
	//  Լ  ʰ core ۽Ű Ҽ  
	//   ̴.
	bool core_init(INIT_CON InitCondition);

	bool core_clock_in(CLK_SIGNAL clock);	// ھ Clock ȣ Էϴ Լ

	//
	// Device Class  ؼ ؾ ϴ Լ
	//
protected:
	//   Լ ݵ Device Ŭ Ǿ Ѵ.
	// frv = file register value
	virtual bool core_get_frv_ext(u8 Bank, u8 RegAddr, u8 *pValue) = NULL;
	virtual	bool core_set_frv_ext(u8 Bank, u8 RegAddr, u8   Value) = NULL;

	//
	//  ھ ϴ    Լ
	//
private:
	bool ExcuteQ1Cycle(void);		// Q1 Clock ھ  ϴ Լ
	bool ExcuteQ2Cycle(void);		// Q2 Clock ھ  ϴ Լ
	bool ExcuteQ3Cycle(void);		// Q3 Clock ھ  ϴ Լ
	bool ExcuteQ4Cycle(void);		// Q4 Clock ھ  ϴ Լ

	bool PopStack(void);			// Stack Pop Ѵ.
	bool PushStack(void);			// Stack Push Ѵ.

	bool SetPC(u16 NewPC);			// PC  Ѵ.
	bool IncPC(void);				// PC = PC + 1

	u16  Fetch(FETCH_CON FetchCon);	// Fetch  ϴ Լ
	bool InstructionDecode(void);	// ɾ ڵ Լ, Q2 Clock .

	//
	//    ϴ Լ
	//
private:
	// BYTE-ORIENTED FILE REGISTER OPERATIONS
	// - ֻ Ʈ ΰ  00 ̴.
	bool Op_ADDWF	(void);
	bool Op_ANDWF	(void);
	bool Op_CLRF	(void);
	bool Op_CLRW	(void);
	bool Op_COMF	(void);
	bool Op_DECF	(void);
	bool Op_DECFSZ	(void);
	bool Op_INCF	(void);
	bool Op_INCFSZ	(void);
	bool Op_IORWF	(void);
	bool Op_MOVF	(void);
	bool Op_MOVWF	(void);
	bool Op_NOP		(void);
	bool Op_RLF		(void);
	bool Op_RRF		(void);
	bool Op_SUBWF	(void);
	bool Op_SWAPF	(void);
	bool Op_XORWF	(void);

	// CONTROL OPERATIONS
	// - ֻ Ʈ ΰ  00  BYTE-ORIENTED  .
	bool Op_CLRWDT	(void);
	bool Op_RETFIE	(void);
	bool Op_RETURN	(void);
	bool Op_SLEEP	(void);

	// BIT-ORIENTED FILE REGISTER OPERATIONS
	// - ֻ Ʈ ΰ  01̴.
	bool Op_BCF		(void);
	bool Op_BSF		(void);
	bool Op_BTFSC	(void);
	bool Op_BTFSS	(void);

	// LITERAL OPERATIONS
	// - ֻ Ʈ ΰ  11̴.
	bool Op_ADDLW	(void);
	bool Op_ANDLW	(void);
	bool Op_IORLW	(void);
	bool Op_MOVLW	(void);
	bool Op_RETLW	(void);
	bool Op_SUBLW	(void);
	bool Op_XORLW	(void);

	// BRANCH OPERATIONS
	// - ֻ Ʈ ΰ  10̴.
	bool Op_CALL	(void);
	bool Op_GOTO	(void);

	//
	// CPIC16xCore Class  ϴ Լ
	//
private:
	// Data Memory   а  Լ ̴.
	// ̵ ڽ ó  óϰ  Լ
	// core_get_frv_ext core_set_frv_ext Լ óѴ.
	bool GetFRV(u8 RegAddr, u8 *pValue);
	bool SetFRV(u8 RegAddr, u8   Value); // FRV : File Register Value

	// Byte-Oriented ɿ   W Register Ǵ File Register
	//  ִ Լ̴.
	bool WriteOpResultToDest(u8 result);

	//
	// CPIC16xCore Class  ϴ 
	//
private:
	bool m_bErrorOccurred;					// ھ Ŭ  ִ üũ
	bool (CPIC16xCore::*m_pOpFunc)(void);	// Decode ɾ Լ  

	Q_CLOCKS m_ClkQ;						// ̹  Q Clock

	u16 m_MemProgramSize;					// ġ Program Memory ũ⸦ Ѵ.
	u16 m_MemEepromSize;					// ġ Eeprom Memory ũ⸦ Ѵ.

	//
	//  ھ Ǵ .
	// Device Class  ϴ.
	//
private:
	//
	// ޸ 
	//
	u16 *m_pMemProgram;		//  14Ʈ
	u8  *m_pMemData;		// 4Page, 8Ʈ
	u8  *m_pMemEeprom;		// 8Ʈ

	//
	//  
	//
	u16 m_RegStack[8];		//  13Ʈ
	u16 m_RegInstruction;	//  14Ʈ
	u8  m_RegWorking;		// 8Ʈ
	s8  m_RegStackPointer;	//  Access   ġ 

	//
	//  ھ Ǵ . (?)
	// CPIC16xCore Class  ϴ.
	//
private:
	u8  m_OpDecodeValue;	// Operation  ϱ  Q2 Clock ڵ 
							// FRV  ӽ÷ ȴ.
};

//
#endif // #ifndef _PIC16X_CORE
